#include <stdio.h>
#include <malloc.h>
#include <string.h>
#ifdef WIN32
#include <io.h>
#include <fcntl.h>
#include <windows.h>
#endif

#define VER     "0.2"

typedef int bool;
#define false   0
#define true    1

// -------------------------------------

long wav_writeheader ( FILE *output, signed __int64 samples )
{
    const long          BitsPerSample       = 16;
    const long          ChannelCount        = 2;
    const long double   SampleFreq          = 44100;
    const long          BytesPerSampleTime  = ((BitsPerSample+7)/8) * ChannelCount;

    unsigned char   header[44];
    unsigned char   *p = header;
    unsigned long   MAXWAVESIZE   = 4294967040LU;
    double          PCMdataLength = (double)BytesPerSampleTime * samples;
    unsigned int    word32;

    *p++ = 'R';
    *p++ = 'I';
    *p++ = 'F';
    *p++ = 'F';                                   // "RIFF" label

    word32 = PCMdataLength + (44 - 8) < (double)MAXWAVESIZE  ?
             (unsigned int)PCMdataLength + (44 - 8)  :  (unsigned int)MAXWAVESIZE;
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);         // Size of the next chunk

    *p++ = 'W';
    *p++ = 'A';
    *p++ = 'V';
    *p++ = 'E';                                   // "WAVE" label

    *p++ = 'f';
    *p++ = 'm';
    *p++ = 't';
    *p++ = ' ';                                   // "fmt " label

    *p++ = 0x10;
    *p++ = 0x00;
    *p++ = 0x00;
    *p++ = 0x00;                                  // length of the PCM data declaration = 2+2+4+4+2+2

    *p++ = 0x01;
    *p++ = 0x00;                                  // ACM type 0x0001 = uncompressed linear PCM

    *p++ = (unsigned char)(ChannelCount >> 0);
    *p++ = (unsigned char)(ChannelCount >> 8);    // Channels

    word32 = (unsigned int) (SampleFreq + 0.5);
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);         // Sample frequency

    word32 *= BytesPerSampleTime;
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);         // Bytes per second in the data stream

    word32 = BytesPerSampleTime;
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);         // Bytes per sample time

    *p++ = (unsigned char)(BitsPerSample >> 0);
    *p++ = (unsigned char)(BitsPerSample >> 8);   // Bits per single sample

    *p++ = 'd';
    *p++ = 'a';
    *p++ = 't';
    *p++ = 'a';                                   // "data" label

    word32 = PCMdataLength < MAXWAVESIZE  ?  (unsigned int)PCMdataLength  :  (unsigned int)MAXWAVESIZE;
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);               // Gre der rohen PCM-Daten

    if ( fwrite (header, 1, p-header, output) != (size_t)(p-header) )
        return -1;

    return p-header;
}

// -------------------------------------

static unsigned long Read32 ( FILE *fp )
{
    unsigned char buff[4];
    if ( fread (buff, 1, 4, fp) != 4 ) return -1;
    return (buff[0] | (buff[1] << 8)) | ((unsigned long)(buff[2] | (buff[3] << 8)) << 16);
}

static unsigned short Read16 ( FILE *fp )
{
    unsigned char buff[2];
    if ( fread (buff, 1, 2, fp) != 2 ) return -1;
    return buff[0] | (buff[1] << 8);
}

// seek to PCM data, return PCM samples
signed __int64 wav_seektopcm ( FILE *fp )
{
    unsigned int Channels;
    unsigned int SampleFreq;
    unsigned int BitsPerSample;
    unsigned int BytesPerSample;
    unsigned int PCMBytes;

    if ( Read32 (fp) != 0x46464952 ) {                  // 4 Byte: check for "RIFF"
        fprintf ( stderr, "Invalid WAV\n" );
        return -1;
    }
    Read32 ( fp );                                      // 4 Byte: chunk size (ignored)
    if ( Read32 (fp) != 0x45564157 ) {                  // 4 Byte: check for "WAVE"
        fprintf ( stderr, "Invalid WAV\n" );
        return -1;
    }
    if ( Read32 (fp) != 0x20746D66 ) {                  // 4 Byte: check for "fmt "
        fprintf ( stderr, "Invalid WAV\n" );
        return -1;
    }
    Read32 (fp);                                        // 4 Byte: read chunk-size (ignored)
    if ( Read16 (fp) != 0x0001 ) {                      // 2 Byte: check for linear PCM
        fprintf ( stderr, "Invalid WAV\n" );
        return -1;
    }
    Channels   = Read16 ( fp );                         // 2 Byte: read no of channels
    SampleFreq = Read32 ( fp );                         // 4 Byte: read sampling frequency
    Read32 ( fp );                                      // 4 Byte: read avg. blocksize (fs*channels*bytepersample)
    Read16 ( fp );                                      // 2 Byte: read byte-alignment (channels*bytepersample)
    BitsPerSample = Read16 ( fp );                      // 2 Byte: read bits per sample
    BytesPerSample = (BitsPerSample + 7) / 8;

    while ( 1 ) {                                       // search for "data"
        if ( feof (fp) ) {
            fprintf ( stderr, "Invalid WAV\n" );
            return -1;
        }
        if ( Read16 (fp) != 0x6164 ) continue;
        if ( Read16 (fp) == 0x6174 ) break;
    }

    PCMBytes = Read32 ( fp );                           // 4 Byte: no. of byte in file
    if ( feof (fp) ) return 0;

    return (signed __int64)(PCMBytes >= 0xFFFFFF00 || PCMBytes == 0 || PCMBytes % (Channels * BytesPerSample) != 0)  ?  (signed __int64)0xFFFFFFFFFFFFFFFF : (signed __int64)PCMBytes / (BytesPerSample * Channels);
}

// -------------------------------------

signed __int64 wav_merge ( FILE *output, const char *fn )
{
    unsigned char   buffer[16364];
    signed __int64  ret;
    FILE            *fp;

    fprintf ( stderr, "Processing '%s'...\n", fn );

    fp = fopen ( fn, "rb" );
    if ( !fp ) {
        fprintf ( stderr, "Failed to open file\n" );
        return -1;
    }

    ret = wav_seektopcm ( fp );
    if ( ret <= 0 ) {
        fclose ( fp );
        return ret;
    }

    while ( !feof (fp) ) {
        size_t bytes = fread ( buffer, 1, sizeof (buffer), fp );
        if ( bytes == 0 ) break;
        fwrite ( buffer, 1, bytes, output );
    }

    fclose ( fp );

    return ret;
}

int main ( int argc, char **argv )
{
    signed __int64      samples  = 0;
    int                 errors   = 0;
    bool                seekable = true;
    FILE                *output;
    int                 i;

    if ( argc < 2 ) {
        fprintf ( stderr, "Wavmerge v" VER "\n" );
        fprintf ( stderr, "Usage: wavmerge <files> [outwav]\n" );
        fprintf ( stderr, "Merge PCM data from all input wavs into outwav, or alternatively to stdout.\n" );
        return 0;
    }

#ifdef WIN32
    {
        static char    **newargv = NULL;
        static int     newargc   = 0;

        newargv = (char **)malloc ( sizeof (*newargv) );
        newargv[newargc++] = strdup ( argv[0] );

        for ( i = 1; i < argc; i++ ) {
            if ( strchr (argv[i], '?') || strchr (argv[i], '*') ) { // wildcards used
                struct _finddata_t  fd;
                long                hFile;

                if ( (hFile = _findfirst (argv[i], &fd)) != -1L ) {
                    if ( !(fd.attrib & _A_SUBDIR || fd.attrib & _A_HIDDEN) ) {  // dirs not included
                        newargv = (char **)realloc ( newargv, (newargc+1) * sizeof (*newargv) );
                        newargv[newargc++] = strdup ( fd.name );
                    }

                    while ( _findnext (hFile, &fd) == 0 ) {
                        if ( !(fd.attrib & _A_SUBDIR || fd.attrib & _A_HIDDEN) ) {  // dirs not included
                            newargv = (char **)realloc ( newargv, (newargc+1) * sizeof (*newargv) );
                            newargv[newargc++] = strdup ( fd.name );
                        }
                    }

                    _findclose ( hFile );
                } else {
                    newargv = (char **)realloc ( newargv, (newargc+1) * sizeof (*newargv) );
                    newargv[newargc++] = strdup ( argv[i] );
                }
            } else {
                newargv = (char **)realloc ( newargv, (newargc+1) * sizeof (*newargv) );
                newargv[newargc++] = strdup ( argv[i] );
            }

            //free ( argv[i] );
        }

        //free ( argv );

        argv = newargv;
        argc = newargc;
    }
#endif

    if ( _access (argv[argc-1], 0) == 0 ) {     // last argument is existing file, output to stdout
        fprintf ( stderr, "Merging WAVs to stdout...\n" );
        output = stdout;
    } else if ( !strcmp (argv[argc-1], "-") ) { // '-' used as alias for stdout
        fprintf ( stderr, "Merging WAVs to stdout...\n" );
        output = stdout;
        argc--;
    } else {                                    // last argument used as name for output file
        fprintf ( stderr, "Merging WAVs to '%s'...\n", argv[argc-1] );
        output = fopen ( argv[argc-1], "wb" );
        if ( !output ) {
            fprintf ( stderr, "Failed to open '%s' for writing\n", argv[argc-1] );
            return 1;
        }
        argc--;
    }

#ifdef WIN32
    if ( output == stdout ) {
        _setmode ( _fileno (stdout), _O_BINARY );
        if ( GetFileType (GetStdHandle (STD_OUTPUT_HANDLE)) != FILE_TYPE_DISK )
            seekable = false;
    }
#endif

    if ( wav_writeheader (output, (unsigned long)-1) < 0 ) {
        fprintf ( stderr, "Failed to write WAV header\n" );
        return 1;
    }

    for ( i = 1; i < argc; i++ ) {
        signed __int64 ret = wav_merge ( output, argv[i] );
        if ( ret >= 0 ) {
            samples += ret;
        } else errors++;
    }

    if ( seekable ) {
        if ( fseek (output, 0, SEEK_SET) == 0 )
            wav_writeheader ( output, samples );
    }

    return errors;
}
